/*
 * Shared Memory definitions & API
 *
 * Copyright (C) 2013-2016 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __TRTL_RT_SMEM_H
#define __TRTL_RT_SMEM_H

/**
 * @defgroup trtl-rt-smem Shared Memory Utilities
 * Collection of utilities to deal with the Mock Turtle Shared Memory.
 * @{
 */

#define TRTL_SMEM_RANGE_ADD 0x10000
#define TRTL_SMEM_RANGE_SUB 0x20000
#define TRTL_SMEM_RANGE_SET 0x30000
#define TRTL_SMEM_RANGE_CLEAR 0x40000
#define TRTL_SMEM_RANGE_FLIP 0x50000
#define TRTL_SMEM_RANGE_TEST_AND_SET 0x60000

/**
 * List of supported Shared Memory operations
 */
enum trtl_smem_type {
	TRTL_SMEM_TYPE_ADD = 1, /**< Atomic addition */
	TRTL_SMEM_TYPE_SUB, /**< Atomic subtraction */
	TRTL_SMEM_TYPE_SET, /**< Atomic bit set */
	TRTL_SMEM_TYPE_CLR, /**< Atomic bit clear */
	TRTL_SMEM_TYPE_FLP, /**< Atomic bit flip */
	TRTL_SMEM_TYPE_TST_SET, /**< Atomic test and set */
};


/**
 * It generates the SHM address range for a given SHM operations
 */
#define TRTL_SMEM_TYPE_TO_RANGE(_type) ((_type) * 0x10000)


/**
 * Compiler attribute to allocate variable on the Share Memory.
 * Add this to the variable declaration to put it on the Shared Memory.
 *
 *     SMEM int x;
 */
#define SMEM volatile __attribute__((section(".smem")))


/**
 * It performs an operations on the shared memory. What the function
 * performs can be summerized as:
 *
 *     (*p) = (*p) <operation-type> x
 *
 * Use this function only when the operation type that you want to use
 * is not supported yet by the library directly. Otherwise,
 * use the other functions.
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 * @param[in] type operation type
 */
static inline void __smem_atomic_op(volatile int *p, int x,
				    enum trtl_smem_type type)
{
	*(volatile int *)(p + (TRTL_SMEM_TYPE_TO_RANGE(type) >> 2)) = x;
}


/**
 * It performs an atomic addition:
 *
 *     (*p) = (*p) + x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_add(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_RANGE_ADD);
}


/**
  * It performs an atomic subtraction:
 *
 *     (*p) = (*p) - x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_sub(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_RANGE_SUB);
}


/**
 * It performs an atomic bitwise OR:
 *
 *     (*p) = (*p) | x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_or(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_RANGE_SET);
}


/**
 * It performs an atomic bitwise "AND ~":
 *
 *     (*p) = (*p) & (~x)
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_and_not(volatile int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_RANGE_CLEAR);
}


/**
 * It performs an atomic bitwise XOR:
 *
 *     (*p) = (*p) ^ x
 *
 * @param[in] p address on the shared memory of the first operator and store location
 * @param[in] x second operation argument
 */
static inline void smem_atomic_xor(int *p, int x)
{
	__smem_atomic_op(p, x, TRTL_SMEM_RANGE_FLIP);
}


/**
 * It does a test and set in the shared memory pointed by p:
 *
 *     val = (*p);
 *     if (val == 0) {
 *        (*p) = 1;
 *     }
 *     return val;
 *
 * This is useful to implement mutex
 *
 * @param[in] p address on the shared memory
 * @return the value before the set
 */
static inline int smem_atomic_test_and_set(int *p)
{
	/* shift right translates range in bytes into range in words */
	return *(volatile int *)(p + (TRTL_SMEM_RANGE_TEST_AND_SET >> 2));
}

/**@}*/

#endif
